#include "prefs-dialog.h"
#include <panel-applet-gconf.h>

#include <string.h>

#include "utils.h"

enum
{
	COLUMN_PERSON,
	COLUMN_UID,
	COLUMN_ICON,
	COLUMN_DISPLAY_NAME,
	NUM_COLUMNS
};

static void mon_person_selected_cb(GtkTreeSelection *sel, PrefsDialog *dialog);

static void
rebuild_stored_person_list(PrefsDialog *dialog)
{
	GSList *slist = NULL;
	GtkTreeIter iter;
	GtkTreeSelection *sel;
	gboolean b;
	GtkTreeModel *model;

	model = GTK_TREE_MODEL(dialog->mon_people_store);

	for (b = gtk_tree_model_get_iter_first(model, &iter);
		 b;
		 b = gtk_tree_model_iter_next(model, &iter))
	{
		const char *uid;

		gtk_tree_model_get(model, &iter,
						   COLUMN_UID, &uid,
						   -1);

		slist = g_slist_append(slist, g_strdup(uid));
	}

	panel_applet_gconf_set_list(dialog->applet->panel_applet,
								PRESENCE_MONITOR_KEY_PEOPLE_MONITORED,
								GCONF_VALUE_STRING, slist, NULL);

	g_slist_foreach(slist, (GFunc)g_free, NULL);
	g_slist_free(slist);

	/* Force an update of the buttons */
	sel = gtk_tree_view_get_selection(
		GTK_TREE_VIEW(dialog->mon_people_treeview));

	mon_person_selected_cb(sel, dialog);
}

static void
mon_person_selected_cb(GtkTreeSelection *sel, PrefsDialog *dialog)
{
	GtkWidget *button;
	GtkTreeIter iter;
	gboolean selected, can_move_up = FALSE, can_move_down = FALSE;

	selected = gtk_tree_selection_get_selected(sel, NULL, &iter);

	if (selected)
	{
		GtkTreePath *path;
		GtkTreeModel *model = GTK_TREE_MODEL(dialog->mon_people_store);

		path = gtk_tree_model_get_path(model, &iter);

		if (path != NULL)
		{
			int *indices = gtk_tree_path_get_indices(path);
			int count    = gtk_tree_model_iter_n_children(model, NULL);
			int index    = indices[0];

			can_move_up   = (index > 0);
			can_move_down = (index < count - 1);
		}
	}

	button = glade_xml_get_widget(dialog->gui, "button-remove-person");
	gtk_widget_set_sensitive(button, selected);

	button = glade_xml_get_widget(dialog->gui, "button-person-up");
	gtk_widget_set_sensitive(button, selected && can_move_up);

	button = glade_xml_get_widget(dialog->gui, "button-person-down");
	gtk_widget_set_sensitive(button, selected && can_move_down);
}

static void
person_selected_cb(GtkTreeSelection *sel, PrefsDialog *dialog)
{
	GtkWidget *button;
	gboolean selected;

	selected = gtk_tree_selection_get_selected(sel, NULL, NULL);

	button = glade_xml_get_widget(dialog->gui, "button-add-person");
	gtk_widget_set_sensitive(button, selected);
}

static void
set_person_info(GtkListStore *store, GtkTreeIter *iter,
				GalagoPerson *person)
{
	GalagoService *service;
	GalagoAccount *account;
	const char *display_name;
	GdkPixbuf *pixbuf = NULL;

	account = galago_person_get_priority_account(person);

	if (account != NULL)
	{
		service = galago_account_get_service(account);

		pixbuf = galago_gdk_pixbuf_new_from_account_with_size(account,
			GTK_ICON_SIZE_MENU);
	}

	display_name = galago_person_get_display_name(person);

	gtk_list_store_set(store, iter,
					   COLUMN_PERSON,       person,
					   COLUMN_UID,          galago_person_get_id(person),
					   COLUMN_ICON,         pixbuf,
					   COLUMN_DISPLAY_NAME, display_name,
					   -1);

	if (pixbuf != NULL)
		g_object_unref(G_OBJECT(pixbuf));
}

static void
add_person_clicked_cb(GtkWidget *widget, PrefsDialog *dialog)
{
	GtkTreeSelection *sel;
	GtkTreeIter src_iter, dest_iter;
	GalagoPerson *person;

	sel = gtk_tree_view_get_selection(GTK_TREE_VIEW(dialog->people_treeview));

	if (!gtk_tree_selection_get_selected(sel, NULL, &src_iter))
		return;

	gtk_tree_model_get(GTK_TREE_MODEL(dialog->people_store), &src_iter,
					   COLUMN_PERSON, &person,
					   -1);

	if (person == NULL)
		return;

	gtk_list_store_append(dialog->mon_people_store, &dest_iter);
	set_person_info(dialog->mon_people_store, &dest_iter, person);

	rebuild_stored_person_list(dialog);
}

static void
remove_person_clicked_cb(GtkWidget *widget, PrefsDialog *dialog)
{
	GtkTreeSelection *sel;
	GtkTreeIter iter;

	sel = gtk_tree_view_get_selection(
		GTK_TREE_VIEW(dialog->mon_people_treeview));

	if (!gtk_tree_selection_get_selected(sel, NULL, &iter))
		return;

	gtk_list_store_remove(dialog->mon_people_store, &iter);

	rebuild_stored_person_list(dialog);
}

static void
move_person_after(GtkListStore *store, GtkTreeIter *iter,
				  GtkTreeIter *position)
{
	GtkTreeIter new_iter;
	GalagoPerson *person;

	gtk_tree_model_get(GTK_TREE_MODEL(store), iter,
					   COLUMN_PERSON, &person,
					   -1);

	gtk_list_store_insert_after(store, &new_iter, position);

	set_person_info(store, &new_iter, person);

	gtk_list_store_remove(store, iter);
}

static void
move_person_before(GtkListStore *store, GtkTreeIter *iter,
				   GtkTreeIter *position)
{
	GtkTreeIter new_iter;
	GalagoPerson *person;

	gtk_tree_model_get(GTK_TREE_MODEL(store), iter,
					   COLUMN_PERSON, &person,
					   -1);

	gtk_list_store_insert_before(store, &new_iter, position);

	set_person_info(store, &new_iter, person);

	gtk_list_store_remove(store, iter);
}

static void
person_up_clicked_cb(GtkWidget *widget, PrefsDialog *dialog)
{
	GtkTreeSelection *sel;
	GtkTreeIter iter, pos_iter;
	GtkTreePath *path;

	sel = gtk_tree_view_get_selection(
		GTK_TREE_VIEW(dialog->mon_people_treeview));

	if (!gtk_tree_selection_get_selected(sel, NULL, &iter))
		return;

	path = gtk_tree_model_get_path(GTK_TREE_MODEL(dialog->mon_people_store),
								   &iter);

	if (!gtk_tree_path_prev(path))
	{
		gtk_tree_path_free(path);

		return;
	}

	if (gtk_tree_model_get_iter(GTK_TREE_MODEL(dialog->mon_people_store),
								&pos_iter, path))
	{
		gtk_list_store_swap(dialog->mon_people_store, &iter, &pos_iter);
	}

	rebuild_stored_person_list(dialog);

	gtk_tree_path_free(path);
}

static void
person_down_clicked_cb(GtkWidget *widget, PrefsDialog *dialog)
{
	GtkTreeSelection *sel;
	GtkTreeIter iter, pos_iter;
	GtkTreePath *path;

	sel = gtk_tree_view_get_selection(
		GTK_TREE_VIEW(dialog->mon_people_treeview));

	if (!gtk_tree_selection_get_selected(sel, NULL, &iter))
		return;

	path = gtk_tree_model_get_path(GTK_TREE_MODEL(dialog->mon_people_store),
								   &iter);

	gtk_tree_path_next(path);

	if (gtk_tree_model_get_iter(GTK_TREE_MODEL(dialog->mon_people_store),
								&pos_iter, path))
	{
		gtk_list_store_swap(dialog->mon_people_store, &iter, &pos_iter);
	}

	rebuild_stored_person_list(dialog);

	gtk_tree_path_free(path);
}

static void
destroy_cb(GtkWidget *window, PrefsDialog *dialog)
{
	const GalagoList *l, *l2, *l2_next;
	GalagoService *service;
	GalagoAccount *account;

	gconf_client_notify_remove(dialog->applet->gconf_client,
							   dialog->show_idle_id);
	gconf_client_notify_remove(dialog->applet->gconf_client,
							   dialog->dim_idle_id);

	for (l = galago_core_get_services(FALSE, FALSE); l != NULL; l = l->next)
	{
		service = (GalagoService *)l->data;

		for (l2 = galago_service_get_accounts(service, FALSE);
			 l2 != NULL;
			 l2 = l2_next)
		{
			account = (GalagoAccount *)l2->data;

			l2_next = l2->next;

			galago_object_unref(account);
		}
	}

	dialog->applet->prefs_dialog = NULL;

	g_free(dialog);
}

static void
mon_people_drag_recv_cb(GtkWidget *treeview, GdkDragContext *ctx,
						  guint x, guint y, GtkSelectionData *sd,
						  guint info, guint t, PrefsDialog *dialog)
{
	if (sd->target == gdk_atom_intern("GALAGO_PERSON", FALSE) &&
		sd->data != NULL)
	{
		GalagoAccount *a = NULL;
		GtkTreePath *path = NULL;
		GtkTreeViewDropPosition position;
		GtkTreeModel *model = GTK_TREE_MODEL(dialog->mon_people_store);

		memcpy(&a, sd->data, sizeof(a));

		if (gtk_tree_view_get_dest_row_at_pos(GTK_TREE_VIEW(treeview), x, y,
											  &path, &position))
		{
			GtkTreeIter iter;
			GalagoAccount *account;
			GValue val = {0};

			gtk_tree_model_get_iter(model,  &iter, path);
			gtk_tree_model_get_value(model, &iter, COLUMN_PERSON, &val);

			account = g_value_get_pointer(&val);

			switch (position)
			{
				case GTK_TREE_VIEW_DROP_AFTER:
				case GTK_TREE_VIEW_DROP_INTO_OR_AFTER:
					move_person_after(dialog->mon_people_store,
									  &dialog->mon_people_drag_iter, &iter);

					break;

				case GTK_TREE_VIEW_DROP_BEFORE:
				case GTK_TREE_VIEW_DROP_INTO_OR_BEFORE:
					move_person_before(dialog->mon_people_store,
									   &dialog->mon_people_drag_iter, &iter);

					break;
			}

			rebuild_stored_person_list(dialog);
		}
	}
}

static void
mon_people_drag_data_cb(GtkWidget *treeview, GdkDragContext *ctx,
						  GtkSelectionData *data, guint info, guint time,
						  PrefsDialog *dialog)
{
	if (data->target == gdk_atom_intern("GALAGO_PERSON", FALSE))
	{
		GtkTreeRowReference *ref;
		GtkTreePath *source_row;
		GtkTreeIter iter;
		GalagoAccount *account = NULL;
		GValue val = {0};

		ref = g_object_get_data(G_OBJECT(ctx), "gtk-tree-view-source-row");
		source_row = gtk_tree_row_reference_get_path(ref);

		if (source_row == NULL)
			return;

		gtk_tree_model_get_iter(GTK_TREE_MODEL(dialog->mon_people_store),
								&iter, source_row);
		gtk_tree_model_get_value(GTK_TREE_MODEL(dialog->mon_people_store),
								 &iter, COLUMN_PERSON, &val);

		dialog->mon_people_drag_iter = iter;

		account = g_value_get_pointer(&val);

		gtk_selection_data_set(data, gdk_atom_intern("GALAGO_PERSON", FALSE),
							   8, (void *)&account, sizeof(account));

		gtk_tree_path_free(source_row);
	}
}

static void
bool_key_changed_cb(GConfClient *client, guint cnxn_id, GConfEntry *entry,
					gpointer widget)
{
	gboolean value = gconf_value_get_bool(gconf_entry_get_value(entry));

	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(widget), value);
}

static void
show_idle_time_toggled_cb(GtkToggleButton *checkbox, PrefsDialog *dialog)
{
	panel_applet_gconf_set_bool(dialog->applet->panel_applet,
								PRESENCE_MONITOR_KEY_SHOW_IDLE_TIME,
								gtk_toggle_button_get_active(checkbox), NULL);
}

static void
dim_idle_toggled_cb(GtkToggleButton *checkbox, PrefsDialog *dialog)
{
	panel_applet_gconf_set_bool(dialog->applet->panel_applet,
								PRESENCE_MONITOR_KEY_DIM_IDLE,
								gtk_toggle_button_get_active(checkbox), NULL);
}

static void
set_values(PrefsDialog *dialog)
{
	GSList *slist, *sl;
	const GalagoList *l;
	GtkTreeIter iter;
	GtkWidget *checkbox;
	char *key;

	/* Set our checkboxes. */
	checkbox = glade_xml_get_widget(dialog->gui, "check-show-idle");
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(checkbox),
		panel_applet_gconf_get_bool(dialog->applet->panel_applet,
									PRESENCE_MONITOR_KEY_SHOW_IDLE_TIME,
									NULL));

	key = panel_applet_gconf_get_full_key(dialog->applet->panel_applet,
										  PRESENCE_MONITOR_KEY_SHOW_IDLE_TIME);
	dialog->show_idle_id =
		gconf_client_notify_add(dialog->applet->gconf_client, key,
								bool_key_changed_cb, checkbox, NULL, NULL);
	g_free(key);

	g_signal_connect(G_OBJECT(checkbox), "toggled",
					 G_CALLBACK(show_idle_time_toggled_cb), dialog);

	checkbox = glade_xml_get_widget(dialog->gui, "check-dim-idle");
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(checkbox),
		panel_applet_gconf_get_bool(dialog->applet->panel_applet,
									PRESENCE_MONITOR_KEY_DIM_IDLE, NULL));

	key = panel_applet_gconf_get_full_key(dialog->applet->panel_applet,
										  PRESENCE_MONITOR_KEY_DIM_IDLE);
	dialog->dim_idle_id =
		gconf_client_notify_add(dialog->applet->gconf_client, key,
								bool_key_changed_cb, checkbox, NULL, NULL);
	g_free(key);

	g_signal_connect(G_OBJECT(checkbox), "toggled",
					 G_CALLBACK(dim_idle_toggled_cb), dialog);

	/* Set our list of monitored accounts. */
	slist = panel_applet_gconf_get_list(dialog->applet->panel_applet,
										PRESENCE_MONITOR_KEY_PEOPLE_MONITORED,
										GCONF_VALUE_STRING, NULL);

	for (sl = slist; sl != NULL; sl = sl->next)
	{
		GalagoPerson *person = NULL;
		GalagoAccount *account = NULL;
		GdkPixbuf *pixbuf = NULL;
		char *str = (char *)sl->data;
		const char *display_name;
		char *account_name = NULL, *service_id = NULL;

		person = galago_core_get_person(str, FALSE, TRUE);

		if (person != NULL)
			account = galago_person_get_priority_account(person);

		if (account != NULL)
		{
			pixbuf = galago_gdk_pixbuf_new_from_account_with_size(account,
				GTK_ICON_SIZE_MENU);

			galago_object_ref(account);
		}

		display_name = galago_person_get_display_name(person);

		gtk_list_store_append(dialog->mon_people_store, &iter);
		gtk_list_store_set(dialog->mon_people_store, &iter,
						   COLUMN_PERSON,       person,
						   COLUMN_UID,          galago_person_get_id(person),
						   COLUMN_ICON,         pixbuf,
						   COLUMN_DISPLAY_NAME, display_name,
						   -1);

		if (pixbuf != NULL)
			g_object_unref(G_OBJECT(pixbuf));

		g_free(str);

		if (service_id != NULL)
			g_free(service_id);

		if (account_name != NULL)
			g_free(account_name);
	}

	g_slist_free(slist);


	/* Set our list of all known people. */
	for (l = galago_core_get_people(FALSE, TRUE); l != NULL; l = l->next)
	{
		const char *display_name = NULL;
		GalagoPerson *person = (GalagoPerson *)l->data;
		GalagoAccount *account;
		GdkPixbuf *pixbuf = NULL;

		account      = galago_person_get_priority_account(person);
		display_name = galago_person_get_display_name(person);

		/*
		 * If NULL, we don't even have any accounts in the list. It's
		 * a completely empty person, and I'm sure it feels saddened by it,
		 * so let's not talk about it too much.
		 */
		if (display_name == NULL)
			continue;

		if (account != NULL)
		{
			pixbuf = galago_gdk_pixbuf_new_from_account_with_size(account,
				GTK_ICON_SIZE_MENU);
		}

		gtk_list_store_append(dialog->people_store, &iter);
		gtk_list_store_set(dialog->people_store, &iter,
						   COLUMN_PERSON,       person,
						   COLUMN_UID,          galago_person_get_id(person),
						   COLUMN_ICON,         pixbuf,
						   COLUMN_DISPLAY_NAME, display_name,
						   -1);

		if (pixbuf != NULL)
			g_object_unref(G_OBJECT(pixbuf));
	}
}

PrefsDialog *
prefs_dialog_new(PresenceApplet *applet)
{
	GladeXML *gui;
	GtkWidget *dialog;
	GtkWidget *button;
	GtkWidget *treeview;
	GtkTreeSelection *sel;
	GtkTreeViewColumn *column;
	GtkCellRenderer *renderer;
	GtkSizeGroup *sg;
	PrefsDialog *prefs_dialog;
	static GtkTargetEntry gte[] = {{"GALAGO_PERSON", GTK_TARGET_SAME_APP, 0}};

	g_return_val_if_fail(applet != NULL, NULL);

	gui = glade_xml_new(PKGDATADIR "/gnome-presence-applet.glade", NULL, NULL);

	dialog = glade_xml_get_widget(gui, "dialog-prefs");

	g_return_val_if_fail(dialog != NULL, NULL);

	prefs_dialog = g_new0(PrefsDialog, 1);

	prefs_dialog->gui    = gui;
	prefs_dialog->dialog = dialog;
	prefs_dialog->applet = applet;

	gtk_window_set_default_size(GTK_WINDOW(dialog), 600, 400);
	gtk_window_set_screen(GTK_WINDOW(dialog),
		gtk_widget_get_screen(GTK_WIDGET(applet->panel_applet)));

	g_signal_connect(G_OBJECT(dialog), "destroy",
					 G_CALLBACK(destroy_cb), prefs_dialog);

	button = glade_xml_get_widget(gui, "button-close");

	g_signal_connect_swapped(G_OBJECT(button), "clicked",
							 G_CALLBACK(gtk_widget_destroy), dialog);


	/* Set up our size group. */
	sg = gtk_size_group_new(GTK_SIZE_GROUP_HORIZONTAL);

	gtk_size_group_add_widget(sg,
		glade_xml_get_widget(gui, "label-monitored-people"));
	gtk_size_group_add_widget(sg,
		glade_xml_get_widget(gui, "label-people"));

	g_object_unref(sg);


	/* Set up the list of monitored people. */
	treeview = glade_xml_get_widget(gui, "treeview-monitored-people");

	prefs_dialog->mon_people_treeview = treeview;

	sel = gtk_tree_view_get_selection(GTK_TREE_VIEW(treeview));

	g_signal_connect(G_OBJECT(sel), "changed",
					 G_CALLBACK(mon_person_selected_cb), prefs_dialog);

	prefs_dialog->mon_people_store =
		gtk_list_store_new(NUM_COLUMNS,
						   G_TYPE_POINTER, G_TYPE_STRING, GDK_TYPE_PIXBUF,
						   G_TYPE_STRING);

	gtk_tree_view_set_model(GTK_TREE_VIEW(treeview),
							GTK_TREE_MODEL(prefs_dialog->mon_people_store));

	column = gtk_tree_view_column_new();
	gtk_tree_view_insert_column(GTK_TREE_VIEW(treeview), column, -1);

	renderer = gtk_cell_renderer_pixbuf_new();
	gtk_tree_view_column_pack_start(column, renderer, FALSE);
	gtk_tree_view_column_add_attribute(column, renderer,
									   "pixbuf", COLUMN_ICON);

	renderer = gtk_cell_renderer_text_new();
	gtk_tree_view_column_pack_start(column, renderer, TRUE);
	gtk_tree_view_column_add_attribute(column, renderer,
									   "text", COLUMN_DISPLAY_NAME);

	gtk_tree_view_enable_model_drag_source(GTK_TREE_VIEW(treeview),
										   GDK_BUTTON1_MASK, gte,
										   1, GDK_ACTION_COPY);
	gtk_tree_view_enable_model_drag_dest(GTK_TREE_VIEW(treeview), gte, 1,
										 GDK_ACTION_COPY | GDK_ACTION_MOVE);

	g_signal_connect(G_OBJECT(treeview), "drag-data-received",
					 G_CALLBACK(mon_people_drag_recv_cb), prefs_dialog);
	g_signal_connect(G_OBJECT(treeview), "drag-data-get",
					 G_CALLBACK(mon_people_drag_data_cb), prefs_dialog);


	/* Set up the list of available accounts. */
	treeview = glade_xml_get_widget(gui, "treeview-people");

	prefs_dialog->people_treeview = treeview;

	sel = gtk_tree_view_get_selection(GTK_TREE_VIEW(treeview));

	g_signal_connect(G_OBJECT(sel), "changed",
					 G_CALLBACK(person_selected_cb), prefs_dialog);

	prefs_dialog->people_store =
		gtk_list_store_new(NUM_COLUMNS,
						   G_TYPE_POINTER, G_TYPE_STRING, GDK_TYPE_PIXBUF,
						   G_TYPE_STRING);

	gtk_tree_view_set_model(GTK_TREE_VIEW(treeview),
							GTK_TREE_MODEL(prefs_dialog->people_store));

	column = gtk_tree_view_column_new();
	gtk_tree_view_insert_column(GTK_TREE_VIEW(treeview), column, -1);

	renderer = gtk_cell_renderer_pixbuf_new();
	gtk_tree_view_column_pack_start(column, renderer, FALSE);
	gtk_tree_view_column_add_attribute(column, renderer,
									   "pixbuf", COLUMN_ICON);

	renderer = gtk_cell_renderer_text_new();
	gtk_tree_view_column_pack_start(column, renderer, TRUE);
	gtk_tree_view_column_add_attribute(column, renderer,
									   "text", COLUMN_DISPLAY_NAME);

	/* Set our values. */
	set_values(prefs_dialog);


	/* Set up our list manipulation buttons. */
	button = glade_xml_get_widget(gui, "button-add-person");

	g_signal_connect(G_OBJECT(button), "clicked",
					 G_CALLBACK(add_person_clicked_cb), prefs_dialog);


	button = glade_xml_get_widget(gui, "button-remove-person");

	g_signal_connect(G_OBJECT(button), "clicked",
					 G_CALLBACK(remove_person_clicked_cb), prefs_dialog);


	button = glade_xml_get_widget(gui, "button-person-up");

	g_signal_connect(G_OBJECT(button), "clicked",
					 G_CALLBACK(person_up_clicked_cb), prefs_dialog);


	button = glade_xml_get_widget(gui, "button-person-down");

	g_signal_connect(G_OBJECT(button), "clicked",
					 G_CALLBACK(person_down_clicked_cb), prefs_dialog);

	return prefs_dialog;
}
